;; Generated package description from excorporate.el
(define-package "excorporate" "0.8.2" "Exchange Web Services (EWS) integration" '((emacs "24.1") (fsm "0.2") (soap-client "3.1.4") (url-http-ntlm "2.0.3") (nadvice "0.3")) :url "http://elpa.gnu.org/packages/excorporate.html" :keywords '("calendar"))
